import json
import tempfile
from pathlib import Path

from .. import Feedstock
from ..stream_capture import redirect_stderr, redirect_stdout
from .base import BaseCommand, common_aliases, common_flags


class ExpandMeta(BaseCommand):
    """
    Application to expand meta.yaml to be fully formed.

    Will execute arbitrary code if necessary to resolve
    dict_object recipes.
    """

    aliases = common_aliases
    flags = common_flags

    def start(self):
        with tempfile.TemporaryDirectory() as d:
            self.fetch(d)
            feedstock = Feedstock(Path(d) / self.feedstock_subdir)
            with redirect_stderr(self.log, {"status": "running"}), redirect_stdout(
                self.log, {"status": "running"}
            ):
                expanded = feedstock.get_expanded_meta()
            if self.json_logs:
                self.log.info(
                    "Expansion complete",
                    extra={"status": "completed", "meta": expanded},
                )
            else:
                self.log.info("Expansion complete\n", extra={"status": "completed"})
                self.log.info(json.dumps(expanded))



from copy import deepcopy
from pathlib import Path

from ruamel.yaml import YAML

yaml = YAML()


class Feedstock:
    """
    A Pangeo Forge feedstock
    """

    def __init__(self, feedstock_dir: Path):
        """
        feedstock_dir: Path to an existing Feedstock repo

        Expects meta.yaml to exist inside in this dir
        """
        self.feedstock_dir = feedstock_dir
        with open(self.feedstock_dir / "meta.yaml") as f:
            self.meta = yaml.load(f)

    def _import(self, spec):
        """
        Import & return given object from recipes/ in feedstock_dir

        spec is of form <filename>:<object-name>, similar to what is used
        elsewhere in python.

        Each file is executed only once and cached.
        """
        if not hasattr(self, "_import_cache"):
            self._import_cache = {}
        module, export = spec.split(":")
        if module not in self._import_cache:
            filename = self.feedstock_dir / f"{module}.py"
            with open(filename) as f:
                ns = {}
                exec(f.read(), ns)
                self._import_cache[module] = ns

        return self._import_cache[module][export]

    def parse_recipes(self):
        """
        Parse the recipes defined in this feedstock & return them

        *Executes arbitrary code* defined in the feedstock recipes.
        """
        recipes = {}
        recipes_config = self.meta.get("recipes")
        if isinstance(recipes_config, list):
            for r in recipes_config:
                recipes[r["id"]] = self._import(r["object"])
        elif isinstance(recipes_config, dict):
            recipes = self._import(recipes_config["dict_object"])
        else:
            raise ValueError("Could not parse recipes config in meta.yaml")

        return recipes

    def get_expanded_meta(self):
        """
        Return full meta.yaml file, expanding recipes if needed.

        recipes are guaranteed to be a list of dictionaries with 'id' keys.
        'object' keys *may* be present, but not guaranteed.
        """
        meta_copy = deepcopy(self.meta)
        if "recipes" in self.meta and "dict_object" in self.meta["recipes"]:
            # We have a dict_object, so let's parse the recipes, and provide
            # keep just the ids, discarding the values - as the values do not
            # actually serialize.
            recipes = self.parse_recipes()
            meta_copy["recipes"] = [{"id": k} for k, v in recipes.items()]

        return meta_copy



"""
Command to run a pangeo-forge recipe
"""
import tempfile
from datetime import datetime
from pathlib import Path

from apache_beam import Pipeline
from pangeo_forge_recipes.storage import StorageConfig
from traitlets import Bool, Type

from .. import Feedstock
from ..bakery.base import Bakery
from ..bakery.local import LocalDirectBakery
from ..storage import InputCacheStorage, MetadataCacheStorage, TargetStorage
from ..stream_capture import redirect_stderr, redirect_stdout
from .base import BaseCommand, common_aliases, common_flags


class Bake(BaseCommand):
    """
    Command to bake a pangeo forge recipe in a given bakery
    """

    aliases = common_aliases
    flags = common_flags | {
        "prune": (
            {"Bake": {"prune": True}},
            "Prune the recipe to run only for 2 time steps",
        )
    }

    prune = Bool(
        False,
        config=True,
        help="""
        Prune the recipe to only run for 2 time steps.

        Makes it much easier to test recipes!
        """,
    )

    bakery_class = Type(
        default_value=LocalDirectBakery,
        klass=Bakery,
        config=True,
        help="""
        The Bakery to bake this recipe in.

        The Bakery (and its configuration) determine which Apache Beam
        Runner is used, and how options for it are specified.
        Defaults to LocalDirectBakery, which bakes the recipe using Apache
        Beam's "DirectRunner". It doesn't use Docker or the cloud, and runs
        everything locally. Useful only for testing!
        """,
    )

    def start(self):
        """
        Start the baking process
        """
        # Create our storage configurations. Traitlets will do its magic, populate these
        # with appropriate config from config file / commandline / defaults.
        target_storage = TargetStorage(parent=self)
        input_cache_storage = InputCacheStorage(parent=self)
        metadata_cache_storage = MetadataCacheStorage(parent=self)

        self.log.info(
            f"Target Storage is {target_storage}\n", extra={"status": "setup"}
        )
        self.log.info(
            f"Input Cache Storage is {input_cache_storage}\n", extra={"status": "setup"}
        )
        self.log.info(
            f"Metadata Cache Storage is {metadata_cache_storage}\n",
            extra={"status": "setup"},
        )

        # Create a temporary directory where we fetch the feedstock repo and perform all operations
        # FIXME: Support running this on an already existing repository, so users can run it
        # as they develop their feedstock
        with tempfile.TemporaryDirectory() as d:
            self.fetch(d)
            feedstock = Feedstock(Path(d) / self.feedstock_subdir)

            self.log.info("Parsing recipes...", extra={"status": "running"})
            with redirect_stderr(self.log, {"status": "running"}), redirect_stdout(
                self.log, {"status": "running"}
            ):
                recipes = feedstock.parse_recipes()

            if self.prune:
                # Prune all recipes if we're asked to
                recipes = {k: r.copy_pruned() for k, r in recipes.items()}

            bakery: Bakery = self.bakery_class(parent=self)

            for name, recipe in recipes.items():
                # Unique name for running this particular recipe.
                # FIXME: Should include the name of repo / ref as well somehow
                job_name = (
                    f"{name}-{recipe.sha256().hex()}-{int(datetime.now().timestamp())}"
                )

                recipe.storage_config = StorageConfig(
                    target_storage.get_forge_target(job_name=job_name),
                    input_cache_storage.get_forge_target(job_name=job_name),
                    metadata_cache_storage.get_forge_target(job_name=job_name),
                )

                pipeline_options = bakery.get_pipeline_options(
                    job_name=job_name,
                    # FIXME: Bring this in from meta.yaml?
                    container_image="pangeo/forge:8a862dc",
                )

                # Set argv explicitly to empty so Apache Beam doesn't try to parse the commandline
                # for pipeline options - we have traitlets doing that for us.
                pipeline = Pipeline(options=pipeline_options, argv=[])
                # Chain our recipe to the pipeline. This mutates the `pipeline` object!
                pipeline | recipe.to_beam()

                # Some bakeries are blocking - if Beam is configured to use them, calling
                # pipeline.run() blocks. Some are not. We handle that here, and provide
                # appropriate feedback to the user too.
                if bakery.blocking:
                    self.log.info(
                        f"Running job for recipe {name}\n",
                        extra={"recipe": "name", "status": "running"},
                    )
                    pipeline.run()
                else:
                    result = pipeline.run()
                    job_id = result.job_id()
                    self.log.info(
                        f"Submitted job {job_id} for recipe {name}",
                        extra={"job_id": job_id, "recipe": name, "status": "submitted"},
                    )



import contextlib
import logging
import sys


class LoggingStream:
    """
    File-like object that redirects to a python logger.

    Minimal implementation used only to replace sys.stdout, stderr.
    """

    def __init__(self, log: logging.Logger, level: str, extra: dict):
        self.log = log
        self.level = level
        self.extra = extra

    def write(self, message):
        self.log.log(self.level, message, extra=self.extra)


@contextlib.contextmanager
def redirect_stdout(log: logging.Logger, extra: dict):
    logging_stdout = LoggingStream(log, logging.INFO, extra)
    original_stdout = sys.stdout
    sys.stdout = logging_stdout
    try:
        yield
    finally:
        sys.stdout = original_stdout


@contextlib.contextmanager
def redirect_stderr(log: logging.Logger, extra: dict):
    logging_stderr = LoggingStream(log, logging.INFO, extra)
    original_stderr = sys.stdout
    sys.stderr = logging_stderr
    try:
        yield
    finally:
        sys.stderr = original_stderr



from traitlets.config import Application

from .commands.bake import Bake
from .commands.expand_meta import ExpandMeta


class App(Application):
    """
    Pangeo Forge Runner CLI

    Primariily used to launch various subcommands
    """

    raise_config_file_errors = True

    subcommands = {
        "expand-meta": (ExpandMeta, "Expand meta.yaml of a config file"),
        "bake": (Bake, "Bake a pangeo-forge recipe with a given Bakery"),
    }

    def start(self):
        self.parse_command_line()
        super().start()


def main():
    app = App()
    app.start()



import logging
import sys

from pythonjsonlogger import jsonlogger
from repo2docker import contentproviders
from traitlets import Bool, List, Unicode
from traitlets.config import Application

# Common aliases we want to support in *all* commands
# The key is what the commandline argument should be, and the
# value is the traitlet config it will be translated to
common_aliases = {
    "log-level": "Application.log_level",
    "f": "BaseCommand.config_file",
    "config": "BaseCommand.config_file",
    "repo": "BaseCommand.repo",
    "ref": "BaseCommand.ref",
    "feedstock-subdir": "BaseCommand.feedstock_subdir",
}

# Common flags we want to support in *all* commands.
# The key is the name of the flag, and the value is a tuple
# consisting of a dicitonary with the traitlet config that will be
# set, and a helpful description to be printed in the commandline
common_flags = {"json": ({"BaseCommand": {"json_logs": True}}, "Generate JSON output")}


class BaseCommand(Application):
    """
    Base Application for all our subcommands.

    Provides common traitlets everyone needs, and base methods for
    fetching a given repository.

    Do not directly instantiate!
    """

    log_level = logging.INFO

    repo = Unicode(
        "",
        config=True,
        help="""
        URL of feedstock repo to operate on.

        Can be anything that is interpretable by self.content_providers,
        using Repo2Docker ContentProviders. By default, this includes Git
        repos, Mercurial Repos, Zenodo, Figshare, Dataverse, Hydroshare,
        Swhid and local file paths.
        """,
    )

    ref = Unicode(
        None,
        allow_none=True,
        config=True,
        help="""
        Ref of feedstock repo to fetch.

        Optional, only used for some methods of fetching (such as git or
        mercurial)
        """,
    )

    feedstock_subdir = Unicode(
        "feedstock",
        config=True,
        help="""
        Subdirectory inside the repository containing the `meta.yaml` file
        """,
    )

    config_file = Unicode(
        "pangeo_forge_runner_config.py",
        config=True,
        help="""
        Load traitlet config from this file if it exists
        """,
    )

    # Content providers from repo2docker are *solely* used to check out a repo
    # and get their contents locally, so we can work on them.
    content_providers = List(
        None,
        [
            contentproviders.Local,
            contentproviders.Zenodo,
            contentproviders.Figshare,
            contentproviders.Dataverse,
            contentproviders.Hydroshare,
            contentproviders.Swhid,
            contentproviders.Mercurial,
            contentproviders.Git,
        ],
        config=True,
        help="""
        List of ContentProviders to use to fetch repo.

        Uses ContentProviders from repo2docker for doing most of the work.
        The ordering matters, and Git is used as the default for any URL
        that we can not otherwise determine.

        If we want to support additional contentproviders, ideally we can
        contribute them upstream to repo2docker.
        """,
    )

    json_logs = Bool(
        False,
        config=True,
        help="""
        Provide JSON formatted logging output to stdout.

        If set to True, *all* output will be emitted as one JSON object per
        line.

        Each line *will* have at least a 'status' field and a 'message' field.
        Various other keys will also be present based on the command being called
        and the value of 'status'.

        TODO: This *must* get a JSON schema.
        """,
    )

    def fetch(self, target_path):
        """
        Fetch repo from url at ref, and check it out to checkout_path

        Uses repo2docker to detect what kinda url is going to be checked out,
        and fetches it into checkout_path. No image building or anything is
        performed.

        checkout_path should be empty.
        """
        picked_content_provider = None
        for ContentProvider in self.content_providers:
            cp = ContentProvider()
            spec = cp.detect(self.repo, ref=self.ref)
            if spec is not None:
                picked_content_provider = cp
                self.log.info(
                    "Picked {cp} content "
                    "provider.\n".format(cp=cp.__class__.__name__),
                    extra={"status": "fetching"},
                )
                break

        if picked_content_provider is None:
            raise ValueError(
                f"Could not fetch {self.repo}, no matching contentprovider found"
            )

        for log_line in picked_content_provider.fetch(
            spec, target_path, yield_output=True
        ):
            self.log.info(log_line, extra=dict(status="fetching"))

    def json_excepthook(self, etype, evalue, traceback):
        """
        Called on an uncaught exception when using json logging

        Avoids non-JSON output on errors when using --json-logs
        """
        self.log.error(
            "Error during running: %s",
            evalue,
            exc_info=(etype, evalue, traceback),
            extra=dict(status="failed"),
        )

    def initialize(self, argv=None):
        super().initialize(argv)
        # Load traitlets config from a config file if present
        self.load_config_file(self.config_file)

        # The application communicates with the outside world via
        # stdout, and we structure this communication via logging.
        # So let's setup the default logger to log to stdout, rather
        # than stderr.
        logHandler = logging.StreamHandler(sys.stdout)

        # Calling logging.getLogger gives us the *root* logger, which we
        # then futz with. Ideally, we'll call getLogger(__name__) which
        # will give us a scoped logger. Unfortunately, apache_beam doesn't
        # do this correctly and fucks with the root logger, and so must we
        # if we want to be able to control all stdout from our CLI (to be JSON or
        # otherwise). FIXME: No extra comes out here, just message
        self.log = logging.getLogger()

        # Remove all existing handlers so we don't repeat messages
        self.log.handlers = []
        self.log.addHandler(logHandler)
        self.log.setLevel(self.log_level)
        # Don't propagate these to the root logger - Apache Beam fucks with the root logger,
        # and we don't want duplicates
        self.log.propagate = False

        # Capture all warnings as well, and route them to our logger
        # This makes sure we don't accidentally write warnings to stderr
        # when calling this with --json
        # FIXME: Some extras need to be set here
        warnings_logger = logging.getLogger("py.warnings")
        warnings_logger.parent = self.log
        logging.captureWarnings(True)

        if self.json_logs:
            # register JSON excepthook to avoid non-JSON output on uncaught exception
            sys.excepthook = self.json_excepthook
            formatter = jsonlogger.JsonFormatter()
            logHandler.setFormatter(formatter)
        else:
            # Since we also have JSON logging, we put newlines in our
            # messages wherever explicitly needed. Avoid the logger
            # adding its own, so we don't have unnecessary blank lines
            logHandler.terminator = ""
            # Just put out the message here, nothing else.
            logHandler.formatter = logging.Formatter(fmt="%(message)s")



from apache_beam.pipeline import PipelineOptions
from traitlets import Bool
from traitlets.config import LoggingConfigurable


class Bakery(LoggingConfigurable):
    """
    Base class for Bakeries where recipes can be run.

    A Bakery provides an opinionated and consistent wrapper to an
    Apache Beam runner. Users only configure what is important to them,
    and the Bakery takes care of the rest.
    """

    blocking = Bool(
        False,
        config=False,
        help="""
        Set to True if this Bakery will default block calls to pipeline.run()

        Not configurable, should be overriden only by subclasses.
        """,
    )

    def get_pipeline_options(
        self, job_name: str, container_image: str
    ) -> PipelineOptions:
        """
        Return a PipelineOptions object that will configure a Pipeline to run on this Bakery

        job_name: A unique string representing this particular run
        container_image: A docker image spec that should be used in this run,
                         if the Bakery supports using docker images

        """
        raise NotImplementedError("Override get_pipeline_options in subclass")



from setuptools import find_packages, setup

with open("README.md") as f:
    readme = f.read()

setup(
    name="pangeo-forge-runner",
    description="Commandline tool to manage pangeo-forge feedstocks",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Yuvi Panda",
    author_email="yuvipanda@gmail.com",
    version="0.4",
    packages=find_packages(),
    install_requires=[
        "jupyter-repo2docker",
        "ruamel.yaml",
        "pangeo-forge-recipes",
        "traitlets",
        "apache-beam[gcp]",
    ],
    entry_points={
        "console_scripts": ["pangeo-forge-runner=pangeo_forge_runner.cli:main"]
    },
)



"""
Bakery for baking pangeo-forge recipes in Direct Runner
"""
from apache_beam.pipeline import PipelineOptions
from traitlets import Integer

from .base import Bakery


class LocalDirectBakery(Bakery):
    """
    Bake recipes on your local machine, without docker.

    Uses the Apache Beam DirectRunner
    """

    # DirectRunner blocks the pipeline.run() call until run is completed
    blocking = True

    num_workers = Integer(
        0,
        config=True,
        help="""
        Number of workers to use when baking the recipe.

        Defaults to 0, which is interpreted by Apache beam to be the
        number of CPUs on the machine
        """,
    )

    def get_pipeline_options(
        self, job_name: str, container_image: str
    ) -> PipelineOptions:
        """
        Return PipelineOptions for use with this Bakery
        """
        # Set flags explicitly to empty so Apache Beam doesn't try to parse the commandline
        # for pipeline options - we have traitlets doing that for us.
        return PipelineOptions(
            flags=[],
            runner="DirectRunner",
            direct_running_mode="multiprocessing",
            direct_num_workers=self.num_workers,
            save_main_session=True,
            # this might solve serialization issues; cf. https://beam.apache.org/blog/beam-2.36.0/
            pickle_library="cloudpickle",
        )



"""
Bakery for baking pangeo-forge recipes in GCP DataFlow
"""
import shutil
import subprocess

from apache_beam.pipeline import PipelineOptions
from traitlets import Bool, TraitError, Unicode, default, validate

from .base import Bakery


class DataflowBakery(Bakery):
    """
    Bake a Pangeo Forge recipe on GCP Dataflow
    """

    project_id = Unicode(
        None,
        allow_none=True,
        config=True,
        help="""
        GCP Project to submit the Dataflow job into.

        Defaults to the output of `gcloud config get-value project` if unset.
        Must be set for the Bakery to function.
        """,
    )

    @default("project_id")
    def _default_project_id(self):
        """
        Set default project_id from `gcloud` if it is set
        """
        if not shutil.which("gcloud"):
            # If `gcloud` is not installed, just do nothing
            return None
        return subprocess.check_output(
            ["gcloud", "config", "get-value", "project"], encoding="utf-8"
        ).strip()

    region = Unicode(
        "us-central1",
        config=True,
        help="""
        GCP Region to submit the Dataflow jobs into
        """,
    )

    machine_type = Unicode(
        "n1-highmem-2",
        config=True,
        help="""
        GCP Machine type to use for the Dataflow jobs
        """,
    )

    use_public_ips = Bool(
        False,
        config=True,
        help="""
        Use public IPs for the Dataflow workers.

        Set to false for projects that have policies against VM
        instances having their own public IPs
        """,
    )

    temp_gcs_location = Unicode(
        None,
        allow_none=True,
        config=True,
        help="""
        GCS URL under which to put temporary files required to launch dataflow jobs

        *Must* be set, and be a gs:// URL.
        """,
    )

    @validate("temp_gcs_location")
    def _validate_temp_gcs_location(self, proposal):
        """
        Ensure that temp_gcs_location is a gs:// URL
        """
        if not proposal["value"].startswith("gs://"):
            raise TraitError("DataflowBakery.temp_gcs_location must be a gs:// URL")
        return proposal["value"]

    def get_pipeline_options(
        self, job_name: str, container_image: str
    ) -> PipelineOptions:
        """
        Return PipelineOptions for use with this Bakery
        """
        if self.temp_gcs_location is None:
            raise ValueError("DataflowBakery.temp_bucket must be set")
        if self.project_id is None:
            raise ValueError("DataflowBakery.project_id must be set")

        # Set flags explicitly to empty so Apache Beam doesn't try to parse the commandline
        # for pipeline options - we have traitlets doing that for us.
        return PipelineOptions(
            flags=[],
            runner="DataflowRunner",
            project=self.project_id,
            job_name=job_name,
            temp_location=self.temp_gcs_location,
            use_public_ips=self.use_public_ips,
            region=self.region,
            # The v2 Runner is required to use custom container images
            # https://cloud.google.com/dataflow/docs/guides/using-custom-containers#usage
            experiments=["use_runner_v2"],
            sdk_container_image=container_image,
            # https://cloud.google.com/dataflow/docs/resources/faq#how_do_i_handle_nameerrors
            save_main_session=True,
            # this might solve serialization issues; cf. https://beam.apache.org/blog/beam-2.36.0/
            pickle_library="cloudpickle",
            machine_type=self.machine_type,
        )



from fsspec import AbstractFileSystem
from pangeo_forge_recipes.storage import CacheFSSpecTarget, FSSpecTarget, MetadataTarget
from traitlets import Dict, Type, Unicode
from traitlets.config import LoggingConfigurable


class StorageTargetConfig(LoggingConfigurable):
    """
    Configuration for Storage Targets when a Pangeo Forge recipe is baked
    """

    fsspec_class = Type(
        klass=AbstractFileSystem,
        config=True,
        help="""
        FSSpec Filesystem to instantiate as class for this target
        """,
    )

    fsspec_args = Dict(
        {},
        config=True,
        help="""
        Args to pass to fsspec_class during instantiation
        """,
    )

    root_path = Unicode(
        "",
        config=True,
        help="""
        Root path under which to put all our storage.

        If {job_name} is present in the root_path, it will be expanded to the
        unique job_name of the current job.
        """,
    )

    pangeo_forge_target_class = Type(
        config=False,
        help="""
        StorageConfig class from pangeo_forge_recipes to instantiate.

        Should be set by subclasses.
        """,
    )

    def get_forge_target(self, job_name: str):
        """
        Return correct pangeo-forge-recipes Target

        If {job_name} is present in `root_path`, it is expanded with the given job_name
        """
        return self.pangeo_forge_target_class(
            self.fsspec_class(**self.fsspec_args),
            root_path=self.root_path.format(job_id=job_name),
        )

    def __str__(self):
        """
        Return sanitized string representation, stripped of possible secrets
        """
        # Only show keys and type of values of args to fsspec, as they might contain secrets
        fsspec_args_filtered = ", ".join(
            f"{k}=<{type(v).__name__}>" for k, v in self.fsspec_args.items()
        )
        return f'{self.pangeo_forge_target_class.__name__}({self.fsspec_class.__name__}({fsspec_args_filtered}, root_path="{self.root_path}")'


class TargetStorage(StorageTargetConfig):
    """
    Storage configuration for where the baked data should be stored
    """

    pangeo_forge_target_class = FSSpecTarget


class InputCacheStorage(StorageTargetConfig):
    """
    Storage configuration for caching input files during recipe baking
    """

    pangeo_forge_target_class = CacheFSSpecTarget


class MetadataCacheStorage(StorageTargetConfig):
    """
    Storage configuration for caching metadata during recipe baking
    """

    pangeo_forge_target_class = MetadataTarget
